import ij.*;
import ij.io.*;
import ij.process.*;
import ij.gui.*;
import ij.text.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import ij.plugin.frame.*;
import java.io.*;
import java.util.*;
import javax.swing.event.*;

/** This plugin implements a framework for spatial convolution
@author Bosco Camillo-BS degree in Computer Science 
        Advisor: Prof. Battiato Sebastiano
        Organization: University of Catania - ITALY
*/
public class Filtering_Framework extends PlugInFrame implements ActionListener, ItemListener
{

        private Font monoFont = new Font("Monospaced", Font.PLAIN, 12);
        private Font sans = new Font("SansSerif", Font.BOLD, 12);
        private Button preview;
        private Button apply;
        private Button viewM;
        private Button insertM;
        private long time;
        private ImageProcessor ip;  
        private ImagePlus imp;
        private int type;
        private Choice choice;
        private String selected;
        private String[] filters={"N-box 3x3","N-box 5x5","Gaussian 3x3","Gaussian 5x5","X-edge","Y-edge","X-Sobel","Y-Sobel","Laplacian","Composite Laplacian"};
        private String kernel1 = "0.111 0.111 0.111\n0.111 0.111 0.111\n0.111 0.111 0.111\n";
        private String kernel1View = "Scale: 9\n1 1 1\n1 1 1\n1 1 1\n";
        private String kernel2 = "0.04 0.04 0.04 0.04 0.04\n0.04 0.04 0.04 0.04 0.04\n0.04 0.04 0.04 0.04 0.04\n0.04 0.04 0.04 0.04 0.04\n0.04 0.04 0.04 0.04 0.04\n";
        private String kernel2View = "Scale: 25\n1 1 1 1 1\n1 1 1 1 1\n1 1 1 1 1\n1 1 1 1 1\n1 1 1 1 1\n";
        private String kernel3 = "0.0625 0.125 0.0625\n0.125 0.25 0.125\n0.0625 0.125 0.0625\n";
        private String kernel3View = "Scale: 16\n1 2 1\n2 4 2\n1 2 1\n";
        private String kernel4 = "0.003 0.015 0.023 0.015 0.003\n0.015 0.0625 0.093 0.0625 0.015\n0.023 0.094 0.141 0.094 0.023\n0.015 0.0625 0.093 0.0625 0.015\n0.003 0.015 0.023 0.015 0.003\n";
        private String kernel4View = "Scale: 256\n1  4   6  4  1\n4 16 24 16 4\n6 24 36 24 6\n4 16 24 16 4\n1  4   6  4  1\n";
        private String kernel5 = "-1 -1 -1\n0 0 0\n1 1 1\n";
        private String kernel6 = "-1 0 1\n-1 0 1\n-1 0 1\n";
        private String kernel7 = "-1 -2 -1\n0 0 0\n1 2 1\n";
        private String kernel8 = "-1 0 1\n-2 0 2\n-1 0 1\n";
        private String kernel9 = "0 -1 0\n-1 4 -1\n0 -1 0\n";
        private String kernel10 = "0 -1 0\n-1 5 -1\n0 -1 0\n";
        
        public Filtering_Framework() 
        {
		super("Filtering Framework");
                setResizable(false);
	}
        public void run(String arg) 
        {
	        
               imp = WindowManager.getCurrentImage();
               
		if (imp==null) 
                {
			IJ.beep();
			IJ.showStatus("No image");
                        IJ.noImage();
			return;
		}
		ip = imp.getProcessor();
                int imgWidth=ip.getWidth();
                int imgHeight=ip.getHeight();
                
                GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		setLayout(gridbag);
                
                // Panel title
                Panel paneltitle = new Panel();
                Label title= new Label("Filtering Framework", Label.CENTER);
		title.setFont(sans);
                paneltitle.add(title);
                c.gridx = 0;
		int y = 0;
		c.gridy = y++;
		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.CENTER;
		c.insets = new Insets(5, 10, 0, 10);//top,left,bottom,right
		gridbag.setConstraints(paneltitle, c);
                add(paneltitle);
                
                Panel panelChoice = new Panel();
                c.gridy = y++;
                c.insets = new Insets(5, 10, 0, 10);
                gridbag.setConstraints(panelChoice, c);
                choice = new Choice();
                for (int i=0; i<filters.length; i++)
                      choice.addItem(filters[i]);
                choice.addItemListener(this);
                selected="N-box 3x3";
                choice.select(0);
                panelChoice.add(choice);
                viewM = new Button("View Mask");
                viewM.addActionListener(this);
                panelChoice.add(viewM);
                add(panelChoice);
                
                //panel Insert Mask
                Panel panelInsertM = new Panel();
                insertM = new Button("Insert Mask & Convolve");
                insertM.addActionListener(this);
                panelInsertM.add(insertM);
                
                // panel Preview
                Panel panelPreview = new Panel();
                preview = new Button("Preview");
                preview.addActionListener(this);
                panelPreview.add(preview);
                
                // panel Apply
                Panel panelApply = new Panel();
                apply = new Button("Apply");
                apply.addActionListener(this);
                panelApply.add(apply);
                
                // panel Info & Help
                Panel panelInfoHelp = new Panel();
                Button ih = new Button("Help & Info");
                ih.addActionListener(new InfoHelpViewerFF());
                panelInfoHelp.add(ih);
                
                // panel container
                Panel container=new Panel();
                container.setLayout(new GridLayout(4,1));
                c.gridy = y++;
                c.insets = new Insets(10, 0, 0, 10);
                gridbag.setConstraints(container, c);
                container.add(panelPreview);
                container.add(panelInsertM);
                container.add(panelApply);
                container.add(panelInfoHelp);
                add(container);
                
                // - - - - - - - - - - - - - - - - - -
                pack();
		GUI.center(this);
		show();
        }
        
        public void actionPerformed(ActionEvent e)
        {
            Object source=e.getSource();
            if(source==viewM)
            {
              if(selected.equals("N-box 3x3"))IJ.showMessage("Mask",kernel1View);
              if(selected.equals("N-box 5x5"))IJ.showMessage("Mask",kernel2View);  
              if(selected.equals("Gaussian 3x3"))IJ.showMessage("Mask",kernel3View);  
              if(selected.equals("Gaussian 5x5"))IJ.showMessage("Mask",kernel4View);  
              if(selected.equals("X-edge"))IJ.showMessage("Mask",kernel5);
              if(selected.equals("Y-edge"))IJ.showMessage("Mask",kernel6);
              if(selected.equals("X-Sobel"))IJ.showMessage("Mask",kernel7);
              if(selected.equals("Y-Sobel"))IJ.showMessage("Mask",kernel8);
              if(selected.equals("Laplacian"))IJ.showMessage("Mask",kernel9);
              if(selected.equals("Composite Laplacian"))IJ.showMessage("Mask",kernel10);
            }
            if(source==insertM)
            {
              boolean scaleOK=false;
              boolean kernelOK=false;
              String scale="";
              String userk="";
              while(!scaleOK&&!kernelOK)
              {
                GenericDialog gd = new GenericDialog("Insert Mask", IJ.getInstance());
                gd.addStringField("Scale: ", "1", 0);
                gd.addTextAreas("", null, 10, 30);
                gd.showDialog();
                if (gd.wasCanceled()) 
                {
		 return;
                }
                scale = gd.getNextString();
                userk = gd.getNextText();
                if(checkScale(scale)&&checkKernel(userk))
                {
                  scaleOK=true;
                  kernelOK=true;
                }
              }
              int scaleVal=Integer.parseInt(scale);
              userk=finalKernel(scaleVal,userk);
              //IJ.showMessage("kernel",userk);
              ImageProcessor processed_ip = ip.crop();
              ImagePlus processed= imp.createImagePlus();
              processed.setProcessor(imp.getTitle(), processed_ip);
              WindowManager.setTempCurrentImage(processed);
              time=System.currentTimeMillis();
              selected=userk;
              IJ.runPlugIn("ConvolverMod_",userk);
              time=System.currentTimeMillis()-time;
              double time_sec=time/1000.0;
              processed.changes=false;
              PreviewImageFF prew=new PreviewImageFF(processed,time_sec);
              WindowManager.setTempCurrentImage(imp);
            }
            if(source==preview)
            {
              ImageProcessor processed_ip = ip.crop();
              ImagePlus processed= imp.createImagePlus();
              processed.setProcessor(imp.getTitle(), processed_ip);
              WindowManager.setTempCurrentImage(processed);
              String chosen="";
              if(selected.equals("N-box 3x3"))chosen=kernel1;
              if(selected.equals("N-box 5x5"))chosen=kernel2;  
              if(selected.equals("Gaussian 3x3"))chosen=kernel3;  
              if(selected.equals("Gaussian 5x5"))chosen=kernel4;  
              if(selected.equals("X-edge"))chosen=kernel5;
              if(selected.equals("Y-edge"))chosen=kernel6;
              if(selected.equals("X-Sobel"))chosen=kernel7;
              if(selected.equals("Y-Sobel"))chosen=kernel8;
              if(selected.equals("Laplacian"))chosen=kernel9;
              if(selected.equals("Composite Laplacian"))chosen=kernel10;
              time=System.currentTimeMillis();
              IJ.runPlugIn("ConvolverMod_",chosen);
              time=System.currentTimeMillis()-time;
              double time_sec=time/1000.0;
              processed.changes=false;
              PreviewImageFF prew=new PreviewImageFF(processed,time_sec);
              WindowManager.setTempCurrentImage(imp);
           }
           if(source==apply)
           {
              Undo.setup(Undo.COMPOUND_FILTER, imp);
              WindowManager.setTempCurrentImage(imp);
              String chosen="";
              if(selected.equals("N-box 3x3"))chosen=kernel1;
              else if(selected.equals("N-box 5x5"))chosen=kernel2;  
                   else if(selected.equals("Gaussian 3x3"))chosen=kernel3;  
                        else if(selected.equals("Gaussian 5x5"))chosen=kernel4;  
                             else if(selected.equals("X-edge"))chosen=kernel5;
                                  else if(selected.equals("Y-edge"))chosen=kernel6;
                                       else if(selected.equals("X-Sobel"))chosen=kernel7;
                                            else if(selected.equals("Y-Sobel"))chosen=kernel8;
                                                else if(selected.equals("Laplacian"))chosen=kernel9;
                                                     else if(selected.equals("Composite Laplacian"))chosen=kernel10;
                                                          else chosen=selected;
              IJ.runPlugIn("ConvolverMod_",chosen);
              imp.updateAndRepaintWindow();
              Undo.setup(Undo.COMPOUND_FILTER_DONE, imp);
              close();
           }
        }
        
        public void itemStateChanged(ItemEvent e)
        {
          selected=choice.getSelectedItem();  
        }
        
        private boolean checkScale(String sc)
        {
          try
          {            
            if(sc.equals(""))
            {
              IJ.error("Error","You must enter a value for scale (default=1)");
              return false;
            }
            int v=Integer.parseInt(sc);
            if(v<=0)
            {
              IJ.error("Error","Scale value must be >= 1");
              return false;
            }
          }            
          catch(NumberFormatException e)
          {
            IJ.error("Error","Scale value must be an integer");
            return false;            
          }
          return true;
        }
        private boolean checkKernel(String k)
        {
            if(k.equals(""))
            {
              IJ.error("Error","You must insert the kernel's values");
              return false;
            }
          StringTokenizer st = new StringTokenizer(k);
          int n = st.countTokens();
          int widthK = (int)Math.sqrt(n);
          if((widthK*widthK)!=n)
          {
            IJ.error("Error","Kernel must be square (nxn)");
            return false;
          }
          
          for (int i=0; i<n; i++)
          {
            String token = st.nextToken();
            try 
            {
              Integer.parseInt(token);
            }
            catch(NumberFormatException e)
            {
              IJ.error("Error","All kernel's values must be integers");
              return false;
            }
          }
          return true; 
        }
        
        private String finalKernel(int scale, String k)
        {
          String output="";
          StringTokenizer st = new StringTokenizer(k);
          int n = st.countTokens();
          int widthK = (int)Math.sqrt(n);
          for (int i=1; i<=n; i++)
          {
            String token = st.nextToken();
            int value=Integer.parseInt(token);
            double v=(double)value/scale;
            if(i%widthK==0)output=output+IJ.d2s(v,3)+"\n";
            else output=output+IJ.d2s(v,3)+" ";
          }
          return output;          
        }
}

class InfoHelpViewerFF implements ActionListener
{
  private String description="HELP ABOUT \"Filtering Framework\"\n\n"+
                                    "This plugin implements a framework for filtering images in the spatial domain.\n"+
                                    "It allows making spatial convolution of the image to filter with a given mask (kernel).\n"+
                                    "User can choose one of the standard filters or insert his/her favourite kernel. This plugin is based on ConvolverMod plugin, a variation of ImageJ's command Convolve....\n"+
                                    "Here is a brief description of the steps to do:\n"+
                                    "1) Choose a standard filter using the combo box;\n"+
                                    "2) You can view the kernel corresponding to the selected standard filter by clicking on the button \"View Mask\";\n"+
                                    "3) Click on the button \"Preview\" to view as the image would be if it was processed according to the chosen filter;\n"+
                                    "4) Also the possibility to enter a different kernel and view the corresponding convolution is provided by clicking on the button \"Insert Mask & Convolve\";\n"+
                                    "5) Click on the button \"Apply\" to definitively apply the current filtering to the image;\n\n"+
                                    "ABOUT THE AUTHOR\n\n"+
                                    "This plugin was implemented by Camillo Bosco under Professor Sebastiano Battiato's supervision.\n"+
                                    "It was a part of a project for the Multimedia Course (MSC Program in Computer Science) at University of Catania (ITALY)\n\n"+
                                    "CONTACT INFO\n\n"+
                                    "For further information contact us:\n"+
                                    "Prof. Battiato Sebastiano, Ph.d. in Computer Science, e-mail: battiato@dmi.unict.it, web: http://www.dmi.unict.it/~battiato\n"+
                                    "Bosco Camillo, BS degree in Computer Science, e-mail: camillo.bosco@studenti.unict.it\n\n"+
                                    "LINKS\n\n"+
                                    "IPLab - Image Processing Laboratory - University of Catania (Italy), web: http://www.dmi.unict.it/~iplab\n"+
                                    "Department of Mathematics and Computer Science - University of Catania (Italy), web: http://www.dmi.unict.it/";
  
  public void actionPerformed(ActionEvent e) 
  {
    StringBuffer sb = new StringBuffer();
    sb.append(description);
    TextWindow tw = new TextWindow("About \" Filtering Framework\"", sb.toString(), 700, 500);
    tw.show();
  }
}

class PreviewImageFF implements ActionListener
{
      private ImageCanvas canvas;
      private Panel panelTime;
      private ImagePlus im;
        
      public PreviewImageFF(ImagePlus img, double time)
      {
        im=img;
        canvas=new ImageCanvas(im);
        ImageWindow imgWin=new ImageWindow(im);
        imgWin.setResizable(false);
        imgWin.setSize(canvas.getSize().width+150,canvas.getSize().height+150);
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints c = new GridBagConstraints();
        imgWin.setLayout(gridbag);
        
        //Panel Canvas
        c.gridx = 0;
        int y=0;
        c.gridy = y++;
        c.insets = new Insets(10, 10, 0, 10);
        gridbag.setConstraints(canvas, c);
        imgWin.add(canvas);
        
        //Panel Time
        Panel panelTime = new Panel();
        c.gridy = y++;
        c.insets = new Insets(2, 10, 0, 10);
        gridbag.setConstraints(panelTime, c);
        Label timeLab = new Label("Elapsed Time: ");
        timeLab.setFont(new Font("SansSerif", Font.PLAIN, 12));
        panelTime.add(timeLab);
        Label timeLab2 = new Label(new Double(time).toString()+" seconds");
        timeLab2.setFont(new Font("SansSerif", Font.PLAIN, 12));
        panelTime.add(timeLab2);
        imgWin.add(panelTime);
        
        Panel panelHisto = new Panel();
        c.gridy = y++;
        c.insets = new Insets(2, 10, 0, 10);
        gridbag.setConstraints(panelHisto, c);
        Button histo = new Button("View Histogram");
        histo.addActionListener(this);
        panelHisto.add(histo);
        imgWin.add(panelHisto);
        imgWin.show();
      }
      
      public void actionPerformed(ActionEvent e) 
      {
        HistogramWindow isto=new HistogramWindow(im.getTitle(),im,256);
        isto.show();
      }
}